﻿
'TODO - ADD watchdog, 

' FOV = 240 degrees
' 10 Hz Update Rate
' 1mm resolution
' 0.36 degrees angular resolution
' 682 Beams 

' This module makes use of only 640 beams and a  FOV of 225.22 degrees


'SCIP is a protocol used to communicate with URG. 
'Commands of SCIP protocols are used to interact with the URG. 

'SCIP commands are used to interact with URG sensor
'SCIP commands are classified as "Sensor information command", "Distance acquisition command" , and "Mode setting command".
'Setting of operation mode
'   "SCIP2.0" ... Change SCIP2.0 mode
'   "TM0", "TM1", "TM2" ... Time stamp mode
'   "SS" ... Change the baudrate
'   "BM" ... laser on.
'   "QT" ... laser off
'   "RS" ... Resets parameter, measurement suspends.
'   "CR" ... Change the rotation speed of the motor.
'Obtain status information
'   "VV" ... Acquisition of version information.
'   "PP" ... Acquisition of parameter information.
'   "II" ... Acquisition of status.
'Receive distance data
'   "GD", "GS" ... Acquisition of distance data one at a time.
'   "MD", "MS" ... Acquisition of distance data continuously.

Imports System.Threading
Imports System.IO.Ports
Imports System.Drawing
Imports System.Windows.Forms


Module modIF_URG
    Friend WithEvents comURG As New SerialPort
    Dim strMessage As String

    Public Function ftnURGInitializeTop() As Boolean
        Try
            ftnURGInitializeTop = False
            boolUrgLaserOn = False
            If Not boolURGConnected And Not boolShutDown And Not boolGetVectorsIsBusy Then
                strStatusDriving = "Detecting URG."
                For i = 1 To 2
                    Application.DoEvents()
                    For Each sp As String In My.Computer.Ports.SerialPortNames
                        ' If Not boolURGConnected And Not boolShutDown And Not ((sp = strBotComPort) And boolBotConnected) _
                        If Not boolURGConnected And Not boolShutDown And Not ((sp = comBot.PortName) And boolBotConnected) _
                            And Not ((sp = ("COM" & frmArm.intArmComPort.ToString)) And frmArm.boolEnableArm) Then   ' 
                            frmDebug.subWriteLogs(" ftnURG_SerialConfig: Checking " & sp)
                            If comURG IsNot Nothing Then ftnUrgClose()
                            comURG.PortName = sp  '
                            ftnURG_CheckConnection()
                            If Not boolURGConnected Then ftnUrgClose()
                        End If
                    Next sp
                Next i
                If boolURGConnected Then
                    frmDebug.subWriteLogs(" ftnURG_SerialConfig: Connected " & comURG.PortName)
                    strStatusDriving = "URG Connected."
                    sngScanFOV = 225.22
                    intScanVectorsMaxRangeInches = 560 / 2.54
                    intScanVectorsMinRangeInches = 0
                    subCalculateScanBeamAngles()
                    subCalculateScanCalCoefficients()
                End If
            ElseIf Not boolGetVectorsIsBusy Then
                boolURGConnected = ftnGetURGScanVectors()
            End If
            strStatusDriving = ""
            ftnURGInitializeTop = boolURGConnected
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftnURGInitializeTop :  " & ex.ToString)
            ftnURGInitializeTop = False
            ftnUrgClose()
        End Try
    End Function

    Public Function ftnURG_SerialOpen() As Boolean
        ' returns true if comBot is successfully opened
        Try
            If Not boolURGConnected Then
                If comURG IsNot Nothing Then
                    If Not comURG.IsOpen Then comURG.Open()
                End If
            End If
            ftnURG_SerialOpen = comURG.IsOpen
            If Not comURG.IsOpen Then frmDebug.subWriteLogs(" ftnURG_SerialOpen: could not open comURG = " & comURG.PortName)
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftnURG_SerialOpen :  " & ex.ToString)
            ftnUrgClose()
        End Try
    End Function

    Function ftnURG_CheckConnection() As Boolean
        Try
            If comURG IsNot Nothing Then
                ftnURG_SerialOpen()
                If comURG.IsOpen Then
                    subURGInitialize()
                    Dim timeGetStart As DateTime = Now()
                    boolURGConnected = ftnGetURGScanVectors()
                    intScannerDelay = Now().Subtract(timeGetStart).TotalMilliseconds
                End If
            Else
                boolURGConnected = False
            End If
            If Not boolURGConnected Then frmDebug.subWriteLogs(" ftnURG_CheckConnection: ERROR  - boolURGConnected = False")
            ftnURG_CheckConnection = boolURGConnected
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftnURG_CheckConnection :  " & ex.ToString)
            ftnUrgClose()
        End Try
    End Function


    Sub subURGInitialize()
        Try
            If Not boolURGConnected Then
                ftnUrgXmit("SCIP2.0", 13, 500) 'protocol
                ftnUrgXmit("RS", 8, 500) ' reset
                boolUrgLaserOn = False
            End If
            If Not boolUrgLaserOn Then
                ftnUrgXmit("BM", 8, 500) ' Laser on
                boolUrgLaserOn = True
            End If
            strMessage = ""
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in subURGInitialize :  " & ex.ToString)
        End Try
    End Sub

    Private Sub subURGDataReceived(sender As System.Object, e As System.IO.Ports.SerialDataReceivedEventArgs) Handles comURG.DataReceived
        Try
            strMessage &= comURG.ReadExisting()
            If boolDebugURG Then frmDebug.subWriteLogs(ftnElapsedTimeS() & " subURGDataReceived:  strMessage =  " & strMessage)
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in subURGDataReceived :  " & ex.ToString)
        End Try
    End Sub

    Function ftnUrgXmit(ByRef strCommand As String, ByVal intLength As Integer, ByVal intTimeOut_ms As Integer) As Boolean
        Try
            ftnUrgXmit = True
            strMessage = ""
            comURG.Write(strCommand & vbCrLf) ' get distance data
            Dim Timeout As DateTime = Now().AddMilliseconds(intTimeOut_ms)
            Do Until ((strMessage.StartsWith(strCommand) And (strMessage.Length >= intLength)) Or boolShutDown Or (Now() > Timeout))
                Thread.Sleep(10)
            Loop
            If (Now() > Timeout) Or (strMessage.Length < intLength) Then
                frmDebug.subWriteLogs(" ftnUrgXmit: ERROR - strCommand  = " & strCommand)
                frmDebug.subWriteLogs(" ftnUrgXmit: ERROR - strMessage  = " & strMessage)
                If (Now() > Timeout) Then frmDebug.subWriteLogs(" ftnUrgXmit: ERROR - timeOut  /  boolURGConnected = " & boolURGConnected)
                If (strMessage.Length < intLength) Then frmDebug.subWriteLogs(" ftnUrgXmit: ERROR -  strMessage.Length = " & strMessage.Length)
                ftnUrgXmit = False
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftnUrgXmit :  " & ex.ToString)
            ftnUrgXmit = False
        End Try
    End Function


    Dim boolGetVectorsIsBusy As Boolean = False
    Dim listData As New List(Of Integer)
    Public Function ftnGetURGScanVectors() As Boolean
        ftnGetURGScanVectors = False
        Try
            If boolDebugURG Then frmDebug.subWriteLogs(ftnElapsedTimeS() & " ftnGetURGScanVectors: START " & tmiMoveInfo.boolBotIsMoving)
            Dim timeOut As DateTime = Now().AddMilliseconds(2000)
            While boolGetVectorsIsBusy And (Now() < timeOut)
                Thread.Sleep(10)
            End While
            If (Now() >= timeOut) Then frmDebug.subWriteLogs(" ftnGetURGScanVectors: ERROR  busy timeOut ")
            boolGetVectorsIsBusy = True
            If comURG.IsOpen Then
                subURGInitialize()
                boolUrgAccessed = True
                If ftnUrgXmit("GD0044072501", 2134, 300) Then ' get distance data
                    Dim strTemp() As String = strMessage.Split(vbLf)
                    Dim strLine As String = ""
                    listData.Clear()
                    For i As Integer = 3 To strTemp.Length - 1
                        If (strTemp(i).Length > 0) Then strLine &= strTemp(i).Substring(0, strTemp(i).Length - 1).ToString()
                    Next
                    listData.AddRange(ftnDecodeStr(strLine))
                    Dim intBeamStep As Integer = constMaxScannerBeams / intNumberOfBeams
                    If boolSlam Then intBeamStep = 1
                    For intBeamNumber As Integer = 0 To constMaxScannerBeams - 1 Step intBeamStep ' Convert to trvScanVectors and apply correction coefficients
                        trvScanVectors(intBeamNumber).sngRange = listData(660 - intBeamNumber)
                        trvScanVectors(intBeamNumber).sngRange = ftnCmToInches(trvScanVectors(intBeamNumber).sngRange) * sngScannerCalCoefficients(intBeamNumber)
                        If boolDebugURG Then frmDebug.subWriteVector(" ftnGetURGScanVectors: trvScanVectors(" & intBeamNumber & ") ", trvScanVectors(intBeamNumber))
                    Next intBeamNumber
                    ftnGetURGScanVectors = True
                Else
                    frmDebug.subWriteLogs(" ftnGetURGScanVectors: ERROR - ftnUrgXmit = false ")
                    subSetBeamsToMax()
                End If
            Else
                frmDebug.subWriteLogs(" ftnGetURGScanVectors: ERROR - comURG is not open ")
                subSetBeamsToMax()
            End If
            boolGetVectorsIsBusy = False
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftnGetURGScanVectors :  " & ex.ToString)
            frmDebug.subWriteLogs("   ftnGetURGScanVectors: strMessage =  " & strMessage)
            boolGetVectorsIsBusy = False
            ftnUrgClose()
            subSetBeamsToMax()
        End Try
    End Function

    Sub subSetBeamsToMax()
        Try
            For intBeamNumber As Integer = 0 To constMaxScannerBeams - 1
                trvScanVectors(intBeamNumber).sngRange = intScanVectorsMaxRangeInches
            Next intBeamNumber
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in subSetBeamsToMax :  " & ex.ToString)
        End Try
    End Sub



    Function ftnDecodeStr(strMessageLine As String) As List(Of Integer)
        Dim listDecode As New List(Of Integer)
        Try
            Dim strTemp As String = ""
            For i As Integer = 0 To strMessageLine.Length Step 3
                If i < strMessageLine.Length - 3 Then
                    strTemp = strMessageLine.Substring(i, 3)
                Else
                    strTemp = strMessageLine.Substring(i)
                End If
                listDecode.Add(ftnDecode(strTemp))
            Next
            If (strMessageLine.Length < (660 * 3)) Then ' ERROR- all data not received
                frmDebug.subWriteLogs(" ftnDecodeStr: ERROR ERROR ERROR Incomplete URG data received " & strMessageLine.Length)
                For i As Integer = strMessageLine.Length To 663 Step 3
                    listDecode.Add(0)
                Next
            End If
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftnDecodeStr :  " & ex.ToString)
        End Try
        Return listDecode
    End Function

    Function ftnDecode(strSixBitEncode As String) As Integer
        Dim intDecode As Integer = 0
        Try
            For Each c As Char In strSixBitEncode
                intDecode <<= 6
                intDecode = intDecode And (Not 63)
                intDecode = intDecode Or (Asc(c) - 48)
            Next
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftnDecode :  " & ex.ToString)
        End Try
        Return intDecode
    End Function

    Public Sub subUrgLaserOff()
        Try
            frmDebug.subWriteLogs(ftnElapsedTimeS() & " subUrgLaserOff:  Start")
            If comURG IsNot Nothing Then
                If boolURGConnected And Not boolShutDown And comURG.IsOpen Then
                    comURG.Write("QT" & vbLf) ' Laser off
                End If
            End If
            boolUrgLaserOn = False
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in subUrgLaserOff " & ex.ToString)
        End Try
    End Sub

    Public Function ftnUrgClose() As Boolean
        Try
            If comURG IsNot Nothing Then comURG.Close()
            boolURGConnected = False
            ftnUrgClose = True
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in ftnUrgClose :  " & ex.ToString)
            ftnUrgClose = False
        End Try
    End Function



End Module
