﻿Imports System.Math
Imports System.Drawing

Module modScanVectors

    Public Function ftnGetScanVectors() As Boolean
        ' Returns true if scan vectors received
        Try
            ftnGetScanVectors = False
            If Not boolSimulatedMovement Then
                If boolHeroSelected Then
                    intNumberOfBeams = 40   '  
                ElseIf tmiMoveInfo.boolBotIsMoving Then
                    intNumberOfBeams = 16   ' reduce number of vectors while moving or lost for faster localization
                ElseIf boolAdjustPose Then
                    intNumberOfBeams = 64
                Else
                    intNumberOfBeams = 32
                End If
                If boolURGConnected Then ' Get URG04 Depth Readings
                    ftnGetScanVectors = ftnGetURGScanVectors()
                    If boolDebugURG And ftnGetScanVectors Then frmDebug.subWriteLogs(ftnElapsedTimeS() & "   URG Scan vectors received")
                ElseIf boolKinectConnected Then ' Get Kinect Depth Readings
                    ftnGetScanVectors = ftnGetKinectScanVectors()
                    If boolDebugKinect And ftnGetScanVectors Then frmDebug.subWriteLogs(ftnElapsedTimeS() & "  Kinect Scan vectors received")
                ElseIf boolHeroSelected And Not tmiMoveInfo.boolBotIsMoving Then
                    ftnGetScanVectors = ftnGetHeroScanVectors()
                Else
                    frmDebug.subWriteLogs(ftnElapsedTimeS() & "  ERROR: Scan vectors unavailable")
                    If Not tmiMoveInfo.boolBotIsMoving Then
                        ftnSendCommand("play phrase Beep")
                        subWait(200)
                        ftnSendCommand("play phrase Beep")
                    End If
                    strStatusDriving = "Scanner Error - Please Reconnect "
                    subSetBeamsToMax()
                End If
            End If
        Catch ex As Exception
            ftnGetScanVectors = False
            frmDebug.subWriteLogs(" EXCEPTION in ftnGetScanVectors - :  " & ex.ToString)
            subSetBeamsToMax()
        End Try
    End Function

    Public Sub subCalculateScanBeamAngles()
        Try
            Dim sngBeamOrientationFactor As Single = sngScanFOV / constMaxScannerBeams
            For intBeamNumber As Integer = 0 To constMaxScannerBeams - 1
                trvScanVectors(intBeamNumber).sngAngle = (sngScanFOV / 2 - sngBeamOrientationFactor * intBeamNumber) + intScanOffsetDegrees
            Next intBeamNumber
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in subCalculateScanBeamAngles :  " & ex.ToString)
        End Try
    End Sub

    Public Sub subCalculateScanCalCoefficients()
        Try
            Dim sngScannerCalibrateCM As Single = sngScannerCalibrate / 10
            Dim sngScannerCurveAdjusted As Single = sngScannerCurveAdjust / ((sngScanFOV / 2) ^ 1.4)
            For intBeamNumber As Integer = 0 To constMaxScannerBeams - 1
                sngScannerCalCoefficients(intBeamNumber) = sngScannerCalibrateCM * (1 + sngScannerCurveAdjusted * Abs(trvScanVectors(intBeamNumber).sngAngle) ^ 1.4)
            Next intBeamNumber
        Catch ex As Exception
            frmDebug.subWriteLogs(" EXCEPTION in subCalculateScanCalCoefficients :  " & ex.ToString)
        End Try
    End Sub


End Module
